# frozen_string_literal: true

# When dead letter queue is used and we encounter non-recoverable message, we should skip it after
# retries and move the broken message to a separate topic
# Throttling should throttle but not break this flow.

setup_karafka(allow_errors: %w[consumer.consume.error])

class Consumer < Karafka::BaseConsumer
  def consume
    # just a check that we have this api method included in the strategy
    collapsed?

    messages.each do |message|
      raise StandardError if message.offset == 3

      DT[:offsets] << message.offset
    end
  end
end

class DlqConsumer < Karafka::BaseConsumer
  def consume
    messages.each do |message|
      DT[:broken] << [message.offset, message.raw_payload]
    end
  end
end

draw_routes do
  topic DT.topics[0] do
    consumer Consumer
    dead_letter_queue(topic: DT.topics[1], max_retries: 1)
    manual_offset_management(true)
    throttling(
      limit: 95,
      interval: 5_000
    )
    virtual_partitions(
      partitioner: ->(_) { rand(10) }
    )
  end

  topic DT.topics[1] do
    consumer DlqConsumer
  end
end

Karafka.monitor.subscribe('error.occurred') do |event|
  next unless event[:type] == 'consumer.consume.error'

  DT[:errors] << 1
end

elements = DT.uuids(100)
produce_many(DT.topic, elements)

start_karafka_and_wait_until do
  DT[:offsets].uniq.count >= 99 && DT.key?(:broken)
end

# we should not have the message that was failing
assert_equal (0..99).to_a.sort - [3], DT[:offsets].sort.uniq

assert_equal DT[:broken].last.last, elements[3]

assert fetch_next_offset.zero?
